mod messages;

use std::marker::PhantomData;

use futures::{SinkExt, StreamExt};
// use futures::{SinkExt, StreamExt};
use serde::{
    de::{self, Visitor},
    ser::SerializeTuple,
    Deserialize, Serialize,
};
// use tokio_tungstenite::tungstenite::Message as WSMessage;

#[derive(Debug)]
struct Message<T> {
    join_ref: Option<String>,
    r#ref: String,
    topic_name: String,
    event_name: String,
    payload: T,
}

impl<T> Serialize for Message<T>
where
    T: Serialize,
{
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut msg = serializer.serialize_tuple(5)?;
        msg.serialize_element(&self.join_ref)?;
        msg.serialize_element(&self.r#ref)?;
        msg.serialize_element(&self.topic_name)?;
        msg.serialize_element(&self.event_name)?;
        msg.serialize_element(&self.payload)?;
        msg.end()
    }
}

impl<'de, T> Deserialize<'de> for Message<T>
where
    T: Deserialize<'de>,
{
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        struct TupleVisitor<T>(PhantomData<T>);

        impl<'de, T> Visitor<'de> for TupleVisitor<T>
        where
            T: Deserialize<'de>,
        {
            type Value = Message<T>;

            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                formatter.write_str("a sequence of values")
            }

            fn visit_seq<A>(self, mut seq: A) -> Result<Self::Value, A::Error>
            where
                A: serde::de::SeqAccess<'de>,
            {
                Ok(Message {
                    join_ref: seq.next_element()?.ok_or_else(|| {
                        de::Error::custom("missing join_ref from sequence at position 0")
                    })?,
                    r#ref: seq.next_element()?.ok_or_else(|| {
                        de::Error::custom("missing `ref` from sequence at position 1")
                    })?,
                    topic_name: seq.next_element()?.ok_or_else(|| {
                        de::Error::custom("missing `topic_name` from sequence at position 2")
                    })?,
                    event_name: seq.next_element()?.ok_or_else(|| {
                        de::Error::custom("missing `event_name` from sequence at position 3")
                    })?,
                    payload: seq.next_element()?.ok_or_else(|| {
                        de::Error::custom("missing `payload` from sequence at position 4")
                    })?,
                })
            }
        }

        deserializer.deserialize_tuple(5, TupleVisitor(PhantomData))
    }
}

pub async fn testchannel() -> anyhow::Result<()> {
    let (ws_stream, _) =
        tokio_tungstenite::connect_async("ws://localhost:4000/agent/websocket?vsn=2.0.0").await?;

    let (mut send, mut recv) = ws_stream.split();

    let msg = Message {
        join_ref: Some(String::from("0")),
        r#ref: String::from("0"),
        topic_name: String::from("agents:1"),
        event_name: String::from("phx_join"),
        payload: String::from("foo"),
    };

    tokio::spawn(async move {
        while let Some(msg) = recv.next().await {
            let msg = msg.unwrap();
            match msg {
                tokio_tungstenite::tungstenite::Message::Binary(bytes) => {
                    let _ = self::messages::parse_binary_message(&bytes[..]).unwrap();
                }
                tokio_tungstenite::tungstenite::Message::Text(text) => {
                    let _ = self::messages::parse_json_message(&text).unwrap();
                }
                _ => todo!(),
            }
        }
    });

    send.send(tokio_tungstenite::tungstenite::Message::Text(
        serde_json::to_string(&msg).unwrap(),
    ))
    .await?;

    // let msg = serde_json::to_string(&Message {
    //     join_ref: Some(String::from("0")),
    //     r#ref: "1",
    //     topic_name: String::from("agents:1"),
    //     event_name: String::from("shout"),
    //     payload: Some(String::from("foo")),
    // })?;

    // send.send(WSMessage::Text(msg)).await?;

    // let msg = Message::heartbeat().to_json()?;
    // send.send(WSMessage::Text(msg)).await?;

    Ok(())
}
